/*
 * A swapon(8)/swapoff(8) for Linux 0.99.
 * swapon.c,v 1.1.1.1 1993/11/18 08:40:51 jrs Exp
 *
 * 1997-02-xx <Vincent.Renardias@waw.com>
 * - added '-s' (summary option)
 * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
 * - added Native Language Support
 * 1999-03-21 Arnaldo Carvalho de Melo <acme@conectiva.com.br>
 * - fixed strerr(errno) in gettext calls
 * 2001-03-22 Erik Troan <ewt@redhat.com>
 * - added -e option for -a
 * - -a shouldn't try to add swaps that are already enabled
 * 2002-04-14 Jari Ruusu
 * - added encrypted swap support
 */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <mntent.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include "xmalloc.h"
#include "swap_constants.h"
#include "nls.h"
#include "fsprobe.h"
#include "realpath.h"
#include "pathnames.h"
#include "sundries.h"
#include "loop.h"
#include "xstrncpy.h"
#include "sha512.h"

#define PATH_MKSWAP	"/sbin/mkswap"

#ifdef HAVE_SYS_SWAP_H
# include <sys/swap.h>
#endif

#ifndef SWAPON_HAS_TWO_ARGS
/* libc is insane, let's call the kernel */
# include <sys/syscall.h>
# define swapon(path, flags) syscall(SYS_swapon, path, flags)
# define swapoff(path) syscall(SYS_swapoff, path)
#endif

#define streq(s, t)	(strcmp ((s), (t)) == 0)

#define QUIET	1
#define CANONIC	1

int all = 0;
int priority = -1;	/* non-prioritized swap by default */

/* If true, don't complain if the device/file doesn't exist */
int ifexists = 0;

static struct option longswaponopts[] = {
		/* swapon only */
	{ "priority", required_argument, 0, 'p' },
	{ "ifexists", 0, 0, 'e' },
	{ "summary", 0, 0, 's' },
		/* also for swapoff */
	{ "all", 0, 0, 'a' },
	{ "help", 0, 0, 'h' },
	{ "verbose", 0, 0, 'v' },
	{ "version", 0, 0, 'V' },
	{ NULL, 0, 0, 0 }
};

static struct option *longswapoffopts = &longswaponopts[2];

static int cannot_find(const char *special);

#define PRINT_USAGE_SPECIAL(_fp) \
		fprintf(_fp, _( \
	"The <special> parameter:\n" \
	" {-L label | LABEL=label}             LABEL of device to be used\n" \
	" {-U uuid  | UUID=uuid}               UUID of device to be used\n" \
	" <device>                             name of device to be used\n" \
	" <file>                               name of file to be used\n\n"))

static void
swapon_usage(FILE *fp, int n) {
	fprintf(fp, _("\nUsage:\n"
	" %1$s -a [-e] [-v]                  enable all swaps from /etc/fstab\n"
	" %1$s [-p priority] [-v] <special>  enable given swap\n"
	" %1$s -s                            display swap usage summary\n"
	" %1$s -h                            display help\n"
	" %1$s -V                            display version\n\n"), progname);

	PRINT_USAGE_SPECIAL(fp);

	exit(n);
}

static void
swapoff_usage(FILE *fp, int n) {
	fprintf(fp, _("\nUsage:\n"
	" %1$s -a [-v]                      disable all swaps\n"
	" %1$s [-v] <special>               disable given swap\n"
	" %1$s -h                           display help\n"
	" %1$s -V                           display version\n\n"), progname);

	PRINT_USAGE_SPECIAL(fp);

	exit(n);
}

/*
 * contents of /proc/swaps
 */
static int numSwaps;
static char **swapFiles;	/* array of swap file and partition names */

static void
read_proc_swaps(void) {
	FILE *swaps;
	char line[1024];
	char *p, **q;

	numSwaps = 0;
	swapFiles = NULL;

	swaps = fopen(_PATH_PROC_SWAPS, "r");
	if (swaps == NULL)
		return;		/* nothing wrong */

	/* skip the first line */
	if (!fgets(line, sizeof(line), swaps)) {
		fprintf (stderr, _("%s: %s: unexpected file format\n"),
			progname, _PATH_PROC_SWAPS);
		fclose(swaps);
		return;
	}
	while (fgets(line, sizeof(line), swaps)) {
		/*
		 * Cut the line "swap_device  ... more info" after device.
		 * This will fail with names with embedded spaces.
		 */
		for (p = line; *p && *p != ' '; p++);
		*p = 0;

		q = realloc(swapFiles, (numSwaps+1) * sizeof(*swapFiles));
		if (q == NULL)
			break;
		swapFiles = q;

		swapFiles[numSwaps++] = strdup(line);
	}
	fclose(swaps);
}

static int
is_in_proc_swaps(const char *fname) {
	int i;

	for (i = 0; i < numSwaps; i++)
		if (swapFiles[i] && !strcmp(fname, swapFiles[i]))
			return 1;
	return 0;
}

static int
display_summary(void)
{
       FILE *swaps;
       char line[1024] ;

       if ((swaps = fopen(_PATH_PROC_SWAPS, "r")) == NULL) {
               int errsv = errno;
               fprintf(stderr, "%s: %s: %s\n", progname, _PATH_PROC_SWAPS,
			strerror(errsv));
               return -1;
       }

       while (fgets(line, sizeof(line), swaps))
               printf("%s", line);

       fclose(swaps);
       return 0 ;
}

static int
swap_is_suspend(const char *device) {
	const char *type = fsprobe_get_fstype_by_devname(device);

	/* S1SUSPEND/S2SUSPEND =
	 *
	 *   "swsuspend" in libblkid
	 *   "suspend" in libvolume_id
	 */
	if (type && (strcmp(type, "suspend") == 0 ||
			strcmp(type, "swsuspend") == 0))
		return 1;

	return 0;
}

/* calls mkswap */
static int
swap_reinitialize(const char *device) {
	const char *label = fsprobe_get_label_by_devname(device);
	const char *uuid  = fsprobe_get_uuid_by_devname(device);
	pid_t pid;
	int status, ret;
	char *cmd[7];
	int idx=0;

	switch((pid=fork())) {
	case -1: /* fork error */
		fprintf(stderr, _("%s: cannot fork: %s\n"),
			progname, strerror(errno));
		return -1;

	case 0:	/* child */
		cmd[idx++] = PATH_MKSWAP;
		if (label && *label) {
			cmd[idx++] = "-L";
			cmd[idx++] = (char *) label;
		}
		if (uuid && *uuid) {
			cmd[idx++] = "-U";
			cmd[idx++] = (char *) uuid;
		}
		cmd[idx++] = (char *) device;
		cmd[idx++] = NULL;
		execv(cmd[0], cmd);
		perror("execv");
		exit(1); /* error  */

	default: /* parent */
		do {
			if ((ret = waitpid(pid, &status, 0)) < 0
					&& errno == EINTR)
				continue;
			else if (ret < 0) {
				fprintf(stderr, _("%s: waitpid: %s\n"),
					progname, strerror(errno));
				return -1;
			}
		} while (0);

		/* mkswap returns: 0=suss, 1=error */
		if (WIFEXITED(status) && WEXITSTATUS(status)==0)
			return 0; /* ok */
	}
	return -1; /* error */
}

static int
do_swapon(const char *orig_special, int prio, int canonic) {
	int status;
	struct stat st;
	const char *special = orig_special;

	if (verbose)
		printf(_("%s on %s\n"), progname, orig_special);

	if (!canonic) {
		special = fsprobe_get_devname(orig_special);
		if (!special)
			return cannot_find(orig_special);
	}

	if (stat(special, &st) < 0) {
		int errsv = errno;
		fprintf(stderr, _("%s: cannot stat %s: %s\n"),
			progname, special, strerror(errsv));
		return -1;
	}

	/* We have to reinitialize swap with old (=useless) software suspend
	 * data. The problem is that if we don't do it, then we get data
	 * corruption the next time an attempt at unsuspending is made.
	 */
	if (swap_is_suspend(special)) {
		fprintf(stdout, _("%s: %s: software suspend data detected. "
					"Reinitializing the swap.\n"),
			progname, special);
		if (swap_reinitialize(special) < 0)
			return -1;
	}

	/* people generally dislike this warning - now it is printed
	   only when `verbose' is set */
	if (verbose) {
		int permMask = (S_ISBLK(st.st_mode) ? 07007 : 07077);

		if ((st.st_mode & permMask) != 0) {
			fprintf(stderr, _("%s: warning: %s has "
					  "insecure permissions %04o, "
					  "%04o suggested\n"),
				progname, special, st.st_mode & 07777,
				~permMask & 0666);
		}
	}

	/* test for holes by LBT */
	if (S_ISREG(st.st_mode)) {
		if (st.st_blocks * 512 < st.st_size) {
			fprintf(stderr,
				_("%s: Skipping file %s - it appears "
				  "to have holes.\n"),
				progname, special);
			return -1;
		}
	}

	{
		int flags = 0;

#ifdef SWAP_FLAG_PREFER
		if (prio >= 0) {
			if (prio > SWAP_FLAG_PRIO_MASK)
				prio = SWAP_FLAG_PRIO_MASK;
			flags = SWAP_FLAG_PREFER
				| ((prio & SWAP_FLAG_PRIO_MASK)
				   << SWAP_FLAG_PRIO_SHIFT);
		}
#endif
		status = swapon(special, flags);
	}

	if (status < 0) {
		int errsv = errno;
		fprintf(stderr, "%s: %s: %s\n",
			progname, orig_special, strerror(errsv));
	}

	return status;
}

static int
cannot_find(const char *special) {
	fprintf(stderr, _("%s: cannot find the device for %s\n"),
		progname, special);
	return -1;
}

static int
swapon_by_label(const char *label, int prio) {
	const char *special = fsprobe_get_devname_by_label(label);
	return special ? do_swapon(special, prio, CANONIC) : cannot_find(label);
}

static int
swapon_by_uuid(const char *uuid, int prio) {
	const char *special = fsprobe_get_devname_by_uuid(uuid);
	return special ? do_swapon(special, prio, CANONIC) : cannot_find(uuid);
}

static int
do_swapoff(const char *orig_special, int quiet, int canonic) {
        const char *special = orig_special;

	if (verbose)
		printf(_("%s on %s\n"), progname, orig_special);

	if (!canonic) {
		special = fsprobe_get_devname(orig_special);
		if (!special)
			return cannot_find(orig_special);
	}

	if (swapoff(special) == 0)
		return 0;	/* success */

	if (errno == EPERM) {
		fprintf(stderr, _("Not superuser.\n"));
		exit(1);	/* any further swapoffs will also fail */
	}

	if (!quiet || errno == ENOMEM) {
		fprintf(stderr, "%s: %s: %s\n",
			progname, orig_special, strerror(errno));
	}
	return -1;
}

static int
swapoff_by_label(const char *label, int quiet) {
	const char *special = fsprobe_get_devname_by_label(label);
	return special ? do_swapoff(special, quiet, CANONIC) : cannot_find(label);
}

static int
swapoff_by_uuid(const char *uuid, int quiet) {
	const char *special = fsprobe_get_devname_by_uuid(uuid);
	return special ? do_swapoff(special, quiet, CANONIC) : cannot_find(uuid);
}

static int
prepare_encrypted_swap(const char *partition, char *loop, char *encryption)
{
	int x, y, fd, ffd;
	int page_size;
	sha512_context s;
	unsigned char b[4096], multiKeyBits[65][32];
	char *a[10], *apiName;
	struct loop_info64 loopinfo;
	FILE *f;

	/*
	 * Some sanity checks
	 */
	if(strlen(partition) < 1) {
		fprintf(stderr, _("swapon: invalid swap device name\n"));
		return 0;
	}
	if(strlen(loop) < 1) {
		fprintf(stderr, _("swapon: invalid loop device name\n"));
		return 0;
	}
	if(strlen(encryption) < 1) {
		fprintf(stderr, _("swapon: invalid encryption type\n"));
		return 0;
	}

	/*
	 * Abort if loop device does not exist or is already in use
	 */
	if((fd = open(loop, O_RDWR)) == -1) {
		fprintf(stderr, _("swapon: unable to open loop device %s\n"), loop);
		return 0;
	}
	if(is_unused_loop_device(fd) == 0) {
		fprintf(stderr, _("swapon: loop device %s already in use\n"), loop);
		goto errout0;
	}

	/*
	 * Compute SHA-512 over first 40 KB of old swap data. This data
	 * is mostly unknown data encrypted using unknown key. SHA-512 hash
	 * output is then used as entropy for new swap encryption key.
	 */
	if(!(f = fopen(partition, "r+"))) {
		fprintf(stderr, _("swapon: unable to open swap device %s\n"), partition);
		goto errout0;
	}
	page_size = getpagesize();
	fseek(f, (long)page_size, SEEK_SET);
	sha512_init(&s);
	for(x = 0; x < 10; x++) {
		if(fread(&b[0], sizeof(b), 1, f) != 1) break;
		sha512_write(&s, &b[0], sizeof(b));
	}
	sha512_final(&s);

	/*
	 * Overwrite 40 KB of old swap data 20 times so that recovering
	 * SHA-512 output beyond this point is difficult and expensive.
	 */
	for(y = 0; y < 20; y++) {
		int z;
		struct {
			struct timeval tv;
			unsigned char h[64];
			int x,y,z;
		} j;
		if(fseek(f, (long)page_size, SEEK_SET)) break;
		memcpy(&j.h[0], &s.sha_out[0], 64);
		gettimeofday(&j.tv, NULL);
		j.y = y;
		for(x = 0; x < 10; x++) {
			j.x = x;
			for(z = 0; z < sizeof(b); z += 64) {
				j.z = z;
				sha512_hash_buffer((unsigned char *)&j, sizeof(j), &b[z], 64);
			}
			if(fwrite(&b[0], sizeof(b), 1, f) != 1) break;
		}
		memset(&j, 0, sizeof(j));
		if(fflush(f)) break;
		if(fsync(fileno(f))) break;
	}
	fclose(f);

	/*
	 * Use all 512 bits of hash output
	 */
	memcpy(&b[0], &s.sha_out[0], 64);
	memset(&s, 0, sizeof(s));

	/*
	 * Read 32 bytes of random entropy from kernel's random
	 * number generator. This code may be executed early on startup
	 * scripts and amount of random entropy may be non-existent.
	 * SHA-512 of old swap data is used as workaround for missing
	 * entropy in kernel's random number generator.
	 */
	if(!(f = fopen("/dev/urandom", "r"))) {
		fprintf(stderr, _("swapon: unable to open /dev/urandom\n"));
		goto errout0;
	}
	fread(&b[64], 32, 1, f);

	/*
	 * Set up struct loop_info64
	 */
	if((ffd = open(partition, O_RDWR)) < 0) {
		fprintf(stderr, _("swapon: unable to open swap device %s\n"), partition);
		goto errout1;
	}
	memset(&loopinfo, 0, sizeof(loopinfo));
	xstrncpy((char *)loopinfo.lo_file_name, partition, LO_NAME_SIZE);
	loopinfo.lo_encrypt_type = loop_crypt_type(encryption, &loopinfo.lo_encrypt_key_size, &apiName);
	if(loopinfo.lo_encrypt_type <= 1) {
		fprintf(stderr, _("swapon: unsupported swap encryption type %s\n"), encryption);
errout2:
		close(ffd);
errout1:
		fclose(f);
errout0:
		close(fd);
		memset(&loopinfo.lo_encrypt_key[0], 0, sizeof(loopinfo.lo_encrypt_key));
		memset(&multiKeyBits[0][0], 0, sizeof(multiKeyBits));
		return 0;
	}
	loopinfo.lo_offset = page_size;
	/* single-key hash */
	sha512_hash_buffer(&b[0], 64+32, &loopinfo.lo_encrypt_key[0], sizeof(loopinfo.lo_encrypt_key));
	/* multi-key hash */
	x = 0;
	while(x < 65) {
		fread(&b[64+32], 16, 1, f);
		sha512_hash_buffer(&b[0], 64+32+16, &multiKeyBits[x][0], 32);
		x++;
	}	

	/*
	 * Try to set up single-key loop
	 */
	if(ioctl(fd, LOOP_SET_FD, ffd) < 0) {
		fprintf(stderr, _("swapon: LOOP_SET_FD failed\n"));
		goto errout2;
	}
	if ((loopinfo.lo_encrypt_type == 18) || (loop_set_status64_ioctl(fd, &loopinfo) < 0)) {
		if(try_cryptoapi_loop_interface(fd, &loopinfo, apiName) < 0) {
			fprintf(stderr, _("swapon: LOOP_SET_STATUS failed\n"));
			ioctl(fd, LOOP_CLR_FD, 0);
			goto errout2;
		}
	}

	/*
	 * Try to put loop to multi-key v3 or v2 mode.
	 * If this fails, then let it operate in single-key mode.
	 */
	if(ioctl(fd, LOOP_MULTI_KEY_SETUP_V3, &multiKeyBits[0][0]) < 0) {
		ioctl(fd, LOOP_MULTI_KEY_SETUP, &multiKeyBits[0][0]);
	}

	/*
	 * Loop is now set up. Clean up the keys.
	 */
	memset(&loopinfo.lo_encrypt_key[0], 0, sizeof(loopinfo.lo_encrypt_key));
	memset(&multiKeyBits[0][0], 0, sizeof(multiKeyBits));
	close(ffd);
	fclose(f);
	close(fd);

	/*
	 * Write 40 KB of zeroes to loop device. That same data is written
	 * to underlying partition in encrypted form. This is done to guarantee
	 * that next time encrypted swap is initialized, the SHA-512 hash will
	 * be different. And, if encrypted swap data writes over this data, that's
	 * even better.
	 */
	if(!(f = fopen(loop, "r+"))) {
		fprintf(stderr, _("swapon: unable to open loop device %s\n"), loop);
		return 0;
	}
	memset(&b[0], 0, sizeof(b));
	for(x = 0; x < 10; x++) {
		if(fwrite(&b[0], sizeof(b), 1, f) != 1) break;
	}
	fflush(f);
	fsync(fileno(f));
	fclose(f);
	sync();

	/*
	 * Run mkswap on loop device so that kernel understands it as swap.
	 * Redirect stderr to /dev/null and ignore exit value.
	 */
	if(!(x = fork())) {
		if((x = open("/dev/null", O_WRONLY)) >= 0) {
			dup2(x, 2);
			close(x);
		}
		a[0] = "mkswap";
		a[1] = loop;
		a[2] = 0;
		execvp(a[0], &a[0]);
		execv("/sbin/mkswap", &a[0]);
		/* error to stdout, stderr is directed to /dev/null */
		printf(_("swapon: unable to execute mkswap\n"));
		exit(1);
	}
	if(x == -1) {
		fprintf(stderr, _("swapon: fork failed\n"));
		return 0;
	}
	waitpid(x, &y, 0);
	sync();

	return 1;
}

static void
shutdown_encrypted_swap(char *loop)
{
	int fd;
	struct stat statbuf;
	struct loop_info64 loopinfo;
	unsigned char b[32];
	FILE *f;

	if(stat(loop, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)) {
		if((fd = open(loop, O_RDWR)) >= 0) {
			if(!loop_get_status64_ioctl(fd, &loopinfo)) {
				/*
				 * Read 32 bytes of random data from kernel's random
				 * number generator and write that to loop device.
				 * This preserves some of kernel's random entropy
				 * to next activation of encrypted swap on this
				 * partition.
				 */
				if((f = fopen("/dev/urandom", "r")) != NULL) {
					fread(&b[0], 32, 1, f);
					fclose(f);
					write(fd, &b[0], 32);
					fsync(fd);
				}
			}
			close(fd);
		}
		sync();
		if((fd = open(loop, O_RDONLY)) >= 0) {
			if(!loop_get_status64_ioctl(fd, &loopinfo)) {
				ioctl(fd, LOOP_CLR_FD, 0);
			}
			close(fd);
		}
	}
}

static int
swapon_all(void) {
	FILE *fp;
	struct mntent *fstab;
	int status = 0;

	read_proc_swaps();

	fp = setmntent(_PATH_MNTTAB, "r");
	if (fp == NULL) {
		int errsv = errno;
		fprintf(stderr, _("%s: cannot open %s: %s\n"),
			progname, _PATH_MNTTAB, strerror(errsv));
		exit(2);
	}

	while ((fstab = getmntent(fp)) != NULL) {
		const char *special;
		int skip = 0;
		int pri = priority;
		char *opt, *opts;
		char *loop = NULL, *encryption = NULL;

		if (!streq(fstab->mnt_type, MNTTYPE_SWAP))
			continue;

		special = fsprobe_get_devname(fstab->mnt_fsname);
		if (!special)
			continue;

		/* parse mount options; */
		opts = strdup(fstab->mnt_opts);
		if (!opts) {
			fprintf(stderr, "not enough memory");
			exit(1);
		}
		for (opt = strtok(opts, ","); opt != NULL; opt = strtok(NULL, ",")) {
			if (strncmp(opt, "pri=", 4) == 0)
				pri = atoi(opt+4);
			if (strcmp(opt, "noauto") == 0)
				skip = 1;
			if (strncmp(opt, "loop=", 5) == 0)
				loop = opt + 5;
			if (strncmp(opt, "encryption=", 11) == 0)
				encryption = opt + 11;
		}
		if(skip)
			continue;
		if (loop && encryption) {
			if(!is_in_proc_swaps(loop) && (!ifexists || !access(special, R_OK))) {
				if (!prepare_encrypted_swap(special, loop, encryption)) {
					status |= -1;
					continue;
				}
				status |= do_swapon(loop, pri, CANONIC);
			}
			continue;
		}
		if (!is_in_proc_swaps(special) && (!ifexists || !access(special, R_OK))) {
			status |= do_swapon(special, pri, CANONIC);
		}
	}
	endmntent(fp);

	return status;
}

static const char **llist = NULL;
static int llct = 0;
static const char **ulist = NULL;
static int ulct = 0;

static void addl(const char *label) {
	llist = (const char **) xrealloc(llist, (++llct) * sizeof(char *));
	llist[llct-1] = label;
}

static void addu(const char *uuid) {
	ulist = (const char **) xrealloc(ulist, (++ulct) * sizeof(char *));
	ulist[ulct-1] = uuid;
}

static int
main_swapon(int argc, char *argv[]) {
	int status = 0;
	int c, i;

	while ((c = getopt_long(argc, argv, "ahep:svVL:U:",
				longswaponopts, NULL)) != -1) {
		switch (c) {
		case 'a':		/* all */
			++all;
			break;
		case 'h':		/* help */
			swapon_usage(stdout, 0);
			break;
		case 'p':		/* priority */
			priority = atoi(optarg);
			break;
		case 'L':
			addl(optarg);
			break;
		case 'U':
			addu(optarg);
			break;
		case 'e':               /* ifexists */
		        ifexists = 1;
			break;
		case 's':		/* status report */
			status = display_summary();
			exit(status);
		case 'v':		/* be chatty */
			++verbose;
			break;
		case 'V':		/* version */
			printf("%s: (%s)\n", progname, PACKAGE_STRING);
			exit(0);
		case 0:
			break;
		case '?':
		default:
			swapon_usage(stderr, 1);
		}
	}
	argv += optind;

	if (!all && !llct && !ulct && *argv == NULL)
		swapon_usage(stderr, 2);

	if (ifexists && (!all || strcmp(progname, "swapon")))
		swapon_usage(stderr, 1);

	if (all)
		status |= swapon_all();

	for (i = 0; i < llct; i++)
		status |= swapon_by_label(llist[i], priority);

	for (i = 0; i < ulct; i++)
		status |= swapon_by_uuid(ulist[i], priority);

	while (*argv != NULL)
		status |= do_swapon(*argv++, priority, !CANONIC);

	return status;
}

static int
main_swapoff(int argc, char *argv[]) {
	FILE *fp;
	struct mntent *fstab;
	int status = 0;
	int c, i;

	while ((c = getopt_long(argc, argv, "ahvVL:U:",
				 longswapoffopts, NULL)) != -1) {
		switch (c) {
		case 'a':		/* all */
			++all;
			break;
		case 'h':		/* help */
			swapoff_usage(stdout, 0);
			break;
		case 'v':		/* be chatty */
			++verbose;
			break;
		case 'V':		/* version */
			printf("%s (%s)\n", progname, PACKAGE_STRING);
			exit(0);
		case 'L':
			addl(optarg);
			break;
		case 'U':
			addu(optarg);
			break;
		case 0:
			break;
		case '?':
		default:
			swapoff_usage(stderr, 1);
		}
	}
	argv += optind;

	if (!all && !llct && !ulct && *argv == NULL)
		swapoff_usage(stderr, 2);

	/*
	 * swapoff any explicitly given arguments.
	 * Complain in case the swapoff call fails.
	 */
	for (i = 0; i < llct; i++)
		status |= swapoff_by_label(llist[i], !QUIET);

	for (i = 0; i < ulct; i++)
		status |= swapoff_by_uuid(ulist[i], !QUIET);

	while (*argv != NULL)
		status |= do_swapoff(*argv++, !QUIET, !CANONIC);

	if (all) {
		/*
		 * In case /proc/swaps exists, unswap stuff listed there.
		 * We are quiet but report errors in status.
		 * Errors might mean that /proc/swaps
		 * exists as ordinary file, not in procfs.
		 * do_swapoff() exits immediately on EPERM.
		 */
		read_proc_swaps();
		for(i=0; i<numSwaps; i++)
			status |= do_swapoff(swapFiles[i], QUIET, CANONIC);

		/*
		 * Unswap stuff mentioned in /etc/fstab.
		 * Probably it was unmounted already, so errors are not bad.
		 * Doing swapoff -a twice should not give error messages.
		 */
		fp = setmntent(_PATH_MNTTAB, "r");
		if (fp == NULL) {
			int errsv = errno;
			fprintf(stderr, _("%s: cannot open %s: %s\n"),
				progname, _PATH_MNTTAB, strerror(errsv));
			exit(2);
		}
		while ((fstab = getmntent(fp)) != NULL) {
			const char *orig_special = fstab->mnt_fsname;
			const char *special;
			int skip = 0;
			char *opt, *opts;
			char *loop = NULL, *encryption = NULL;

			if (!streq(fstab->mnt_type, MNTTYPE_SWAP))
				continue;

			special = fsprobe_get_devname(orig_special);
			if (!special)
				continue;

			/* parse mount options; */
			opts = strdup(fstab->mnt_opts);
			if (!opts) {
				fprintf(stderr, "not enough memory");
				exit(1);
			}
			for (opt = strtok(opts, ","); opt != NULL; opt = strtok(NULL, ",")) {
				if (strcmp(opt, "noauto") == 0)
					skip = 1;
				if (strncmp(opt, "loop=", 5) == 0)
					loop = opt + 5;
				if (strncmp(opt, "encryption=", 11) == 0)
					encryption = opt + 11;
				if (strncmp(opt, "encrypted", 9) == 0)
					skip=1;
			}
			if (loop && encryption) {
				if (!is_in_proc_swaps(loop)) {  
					if(skip)
						continue;
					do_swapoff(loop, QUIET, CANONIC);
				}
				shutdown_encrypted_swap(loop);
				continue;
			}
			if(skip)
				continue;
			if (!is_in_proc_swaps(special)) {
				do_swapoff(special, QUIET, CANONIC);
			}


		}
		endmntent(fp);
	}

	return status;
}

int
main(int argc, char *argv[]) {
	char *p;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	progname = argv[0];
	p = strrchr(progname, '/');
	if (p)
		progname = p+1;

	if (streq(progname, "swapon"))
		return main_swapon(argc, argv);
	else
		return main_swapoff(argc, argv);
}
